/*
  forward.c

  $Id: forward.c,v 1.6 1999/12/06 01:52:55 everton Exp $
 */

#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <signal.h>
#include "portfwd.h"
#include "forward.h"
#include "util.h"
#include "solve.h"
#include "host_map.hpp"
#include "iterator.hpp"

int grandchild_pid[MAX_FD];

/*
 * Returns -1 on failure; file descriptor on success.
 */
int tcp_connect(const struct ip_addr *ip, int port)
{
  int rsd = socket(AF_INET, SOCK_STREAM, get_protonumber(P_TCP));
  if (rsd == -1) {
    syslog(LOG_ERR, "tcp_connect: Can't create TCP socket: %m");
    return -1;
  }

  struct sockaddr_in sa;
  sa.sin_family      = AF_INET;
  sa.sin_port        = htons(port);
  sa.sin_addr.s_addr = *((unsigned int *) ip->addr);
  memset((char *) sa.sin_zero, 0, sizeof(sa.sin_zero));

  if (connect(rsd, (struct sockaddr *) &sa, sizeof(sa))) {
    syslog(LOG_ERR, "tcp_connect: Can't connect to %s:%d: %m", inet_ntoa(sa.sin_addr), port);
    close(rsd);
    return -1;
  }

  return rsd;
}

/*
 * Returns socket descriptor on success; -1 on failure.
 */
int tcp_match(const vector<host_map*> *map_list, const struct ip_addr *ip, int port)
{
  int rsd;
  iterator<vector<host_map*>,host_map*> it(*map_list);
  for (it.start(); it.cont(); it.next()) {
    rsd = it.get()->tcp_match(ip, port);
    if (rsd != -1)
      return rsd;
  }
  return -1;
}

void m_close(int fd) 
{
  if (shutdown(fd, 0)) 
    syslog(LOG_ERR, "shutdown() on socket failed: %m");
  if (close(fd))
    syslog(LOG_ERR, "close() on socket failed: %m");
}

void quit_handler(int sig)
{
  ONVERBOSE(syslog(LOG_DEBUG, "child: quit_handler: Grandchild with PID %d exiting under request", getpid()));

  exit(0);
}

int simple_buf_copy(int src_fd, int trg_fd)
{
  char buf[BUF_SZ];
  int rd = read(src_fd, buf, BUF_SZ);
  if (!rd)
    return -1;
  if (rd < 0) {
    syslog(LOG_ERR, "simple_copy: Failure reading from socket: %m");
    return -1;
  }

  int wr = write(trg_fd, buf, rd);
  if (!wr)
    return -1;
  if (wr < rd) {
    syslog(LOG_ERR, "simple_copy: Partial write to socket: %m");
    return -1;
  }
  return 0;
}

void simple_tcp_forward(int sd, struct ip_addr *remote_ip, int remote_port)
{
  int dest_fd[MAX_FD];

  fd_set fds, tmp_fds;
  int fd_set_len = sizeof(fd_set);
  FD_ZERO(&fds);
  FD_SET(sd, &fds);
  int maxfd = sd + 1;
  int nd;

  for (;;) { /* forever */
    /*
     * Restores tmp_fds from fds.
     */
    memcpy(&tmp_fds, &fds, fd_set_len);

    /*
     * Wait for event: connection on sd or data on anyone else.
     */
    nd = select(maxfd, &tmp_fds, 0, 0, 0);
    if (nd == -1) {
      syslog(LOG_ERR, "simple_forward: TCP select() failed: %m");
      continue;
    }

    /*
     * Let's handle sd here (mother socket).
     */
    if (FD_ISSET(sd, &tmp_fds)) {

      --nd;

      /*
       * Clear sd so it's ignored later.
       */
      FD_CLR(sd, &tmp_fds); 

      struct sockaddr_in cli_sa;
      unsigned int cli_sa_len = sizeof(cli_sa);
      int csd = accept(sd, (struct sockaddr *) &cli_sa, &cli_sa_len);
      if (csd < 0)
	syslog(LOG_ERR, "simple_forward: Can't accept TCP socket: %m");
      else {
	int cli_port = ntohs(cli_sa.sin_port);
	ONVERBOSE(syslog(LOG_DEBUG, "simple_forward: TCP connection from %s:%d", inet_ntoa(cli_sa.sin_addr), cli_port));

	/*
	 * Connect to destination.
	 */
	  
	int rsd = tcp_connect(remote_ip, remote_port);
	if (rsd != -1) {
	  if ((csd >= MAX_FD) || (rsd >= MAX_FD)) {
	    syslog(LOG_ERR, "simple_forward: Destination socket descriptors overflow");
	    m_close(csd);
	    m_close(rsd);
	  }
	  else {
	    /*
	     * Add pair of communicating sockets.
	     */
	    FD_SET(csd, &fds);
	    FD_SET(rsd, &fds);
	    maxfd = MAX(maxfd, MAX(rsd + 1, csd + 1)); 
	  
	    /*
	     * Save peers so they can be remembered later.
	     */
	    dest_fd[csd] = rsd;
	    dest_fd[rsd] = csd;
	  }
	}
      }

    } /* sd (mother socket) handled */

    /*
     * Now all other sockets.
     */
    for (int src_fd = 0; nd; ++src_fd)
      if (FD_ISSET(src_fd, &tmp_fds)) {
	--nd;
	int trg_fd = dest_fd[src_fd];
	/*
	 * Copy data.
	 */
	int fail = buf_copy(src_fd, trg_fd, 0);
	if (fail) {
	  /*
	   * Remove pair of communicating sockets.
	   */
	  FD_CLR(src_fd, &fds);
	  FD_CLR(trg_fd, &fds);
	  m_close(src_fd);
	  m_close(trg_fd);
	}
      }

  } /* main loop */

}

int ftp_spawn(struct ip_addr *local_ip, int *local_port, struct ip_addr *remote_ip, int remote_port)
{
  int sd = tcp_listen(local_ip, local_port, 3);
  if (sd == -1) {
    syslog(LOG_ERR, "FTP spawn: Can't listen: %m");
    return -1;
  }

  pid_t pid = fork();
  if (pid) {
    if (pid < 0)
      syslog(LOG_ERR, "FTP spawn: fork() failed: %m");

    return pid;
  }

  /* Grandchild */

  signal(SIGCHLD, SIG_IGN);
  signal(SIGUSR1, quit_handler);

  simple_tcp_forward(sd, remote_ip, remote_port);

  syslog(LOG_ERR, "Grandchild exiting (!)");
  exit(1);
}

void gc_clean(int fd)
{
  pid_t pid = grandchild_pid[fd];
  if (pid != -1) {
    ONVERBOSE(syslog(LOG_DEBUG, "Requesting termination of PID %d for FD %d", pid, fd));

    if (kill(pid, SIGUSR1))
      syslog(LOG_ERR, "Can't request grandchild termination for PID: %d: %m", pid);
    grandchild_pid[fd] = -1;
  }
}

void gc_fill(int fd, pid_t pid)
{
  gc_clean(fd);

  grandchild_pid[fd] = pid;

  ONVERBOSE(syslog(LOG_DEBUG, "PID %d stored for termination request on FD %d", pid, fd));
}

int ftp_passive(char *buf, int *rd, int src_fd, int trg_fd)
{
  if (memcmp(buf, "227", 3))
    return 0;

  ONVERBOSE(syslog(LOG_DEBUG, "Passive FTP reply detected: %s", buf));

  char *i = strchr(buf, '(');
  if (!i) {
    syslog(LOG_ERR, "Missing remote address in passive FTP reply");
    return -1;
  }
  ++i;

  int ad[4];
  int port[2];
  
  if (sscanf(i, "%d,%d,%d,%d,%d,%d", 
	    &ad[0], &ad[1], &ad[2], &ad[3],
	    &port[0], &port[1]) < 6) {
    syslog(LOG_ERR, "Passive FTP reply mismatch");
    return -1;
  }

  /*
   * Remote address.
   */

  char addr[4];
  addr[0] = ad[0];
  addr[1] = ad[1];
  addr[2] = ad[2];
  addr[3] = ad[3];

  struct ip_addr remote_ip;
  remote_ip.len = addr_len;
  remote_ip.addr = addr;
  int remote_port = (port[0] << 8) | port[1];

  ONVERBOSE(syslog(LOG_DEBUG, "Remote address extracted: %s:%d", addrtostr(&remote_ip), remote_port));

  /*
   * Local address.
   */

  const size_t NAME_LEN = 32;
  char localname[NAME_LEN];
  
  if (gethostname(localname, NAME_LEN)) {
    syslog(LOG_ERR, "Can't get local hostname: %m");
    return -1;
  }

  struct ip_addr local_ip = solve_hostname(localname);

  const char *local_addr = local_ip.addr;

  /*
   * FTP forwarder.
   */
    
  int local_port = 0;

  pid_t pid = ftp_spawn(&local_ip, &local_port, &remote_ip, remote_port);
  if (pid == -1) {
    syslog(LOG_ERR, "Can't spawn FTP forwarder");
    free(local_ip.addr);
    return -1;
  }

  /*
   * Address rewriting.
   */
  if (snprintf(buf, BUF_SZ, "227 Portfwd Passive Mode (%u,%u,%u,%u,%u,%u)\n", 
	       local_addr[0], 
	       local_addr[1], 
	       local_addr[2], 
	       local_addr[3],
	       (local_port >> 8) & 0xFF, local_port & 0xFF) == -1)
    syslog(LOG_WARNING, "FTP reply truncated");

  free(local_ip.addr);

  *rd = strlen(buf);

  gc_fill(src_fd, pid);
  gc_fill(trg_fd, pid);

  return 0;
}

int buf_copy(int src_fd, int trg_fd, int pasv)
{
  char buf[BUF_SZ];
  int rd = read(src_fd, buf, BUF_SZ);
  if (!rd)
    return -1;
  if (rd < 0) {
    syslog(LOG_ERR, "copy: Failure reading from socket: %m");
    return -1;
  }

  if (pasv)
    if (ftp_passive(buf, &rd, src_fd, trg_fd))
      return -1;

  int wr = write(trg_fd, buf, rd);
  if (!wr)
    return -1;
  if (wr < rd) {
    syslog(LOG_ERR, "copy: Partial write to socket: %m");
    return -1;
  }
  return 0;
}

int drop_privileges(int uid, int gid)
{
  if (gid != -1)
    if (setregid(gid, gid)) {
      syslog(LOG_ERR, "Can't switch to group: %d: %m", gid);
      return -1;
    }

  if (uid != -1)
    if (setreuid(uid, uid)) {
      syslog(LOG_ERR, "Can't switch to user: %d: %m", uid);
      return -1;
    }

  return 0;
}

int tcp_listen(struct ip_addr *ip, int *port, int queue)
{
  int sd = socket(AF_INET, SOCK_STREAM, get_protonumber(P_TCP));
  if (sd == -1) {
    syslog(LOG_ERR, "listen: Can't create TCP socket: %s\n", strerror(errno));
    return -1;
  }
    
  struct sockaddr_in sa;
  unsigned int sa_len = sizeof(sa);
  sa.sin_family      = AF_INET;
  sa.sin_port        = htons(*port);
  sa.sin_addr.s_addr = *((unsigned int *) ip->addr);
  memset((char *) sa.sin_zero, 0, sizeof(sa.sin_zero));

  if (bind(sd, (struct sockaddr *) &sa, sa_len)) {
    syslog(LOG_ERR, "listen: Can't bind TCP socket: %m: %s:%d\n", inet_ntoa(sa.sin_addr), port);
    close(sd);
    return -1;
  }

  if (!*port) {
    if (getsockname(sd, (struct sockaddr *) &sa, &sa_len)) {
      syslog(LOG_ERR, "listen: Can't get local sockname: %m");
      return -1;
    }
    *port = ntohs(sa.sin_port);
  }

  if (listen(sd, queue)) {
    syslog(LOG_ERR, "listen: Can't listen TCP socket: %s\n", strerror(errno));
    close(sd);
    return -1;
  }

  ONVERBOSE(syslog(LOG_DEBUG, "Listening TCP connection on %s:%d", inet_ntoa(sa.sin_addr), *port));

  return sd;
}

void do_tcp_forward(int sd, int port, vector<host_map*> *map_list, int pasv, int uid, int gid)
{
  if (drop_privileges(uid, gid)) {
    close(sd);
    return;
  }

  int dest_fd[MAX_FD];

  for (int fd = 0; fd < MAX_FD; ++fd)
    grandchild_pid[fd] = -1;

  fd_set fds, tmp_fds;
  int fd_set_len = sizeof(fd_set);
  FD_ZERO(&fds);
  FD_SET(sd, &fds);
  int maxfd = sd + 1;
  int nd;

  for (;;) { /* forever */
    /*
     * Restores tmp_fds from fds.
     */
    memcpy(&tmp_fds, &fds, fd_set_len);

    /*
     * Wait for event: connection on sd or data on anyone else.
     */
    nd = select(maxfd, &tmp_fds, 0, 0, 0);
    if (nd == -1) {
      if (errno != EINTR)
	syslog(LOG_ERR, "TCP select() failed: %m");
      continue;
    }

    /*
     * Let's handle sd here (mother socket).
     */
    if (FD_ISSET(sd, &tmp_fds)) {

      --nd;

      /*
       * Clear sd so it's ignored later.
       */
      FD_CLR(sd, &tmp_fds); 

      struct sockaddr_in cli_sa;
      unsigned int cli_sa_len = sizeof(cli_sa);
      int csd = accept(sd, (struct sockaddr *) &cli_sa, &cli_sa_len);
      if (csd < 0)
	syslog(LOG_ERR, "Can't accept TCP socket: %m");
      else {
	int cli_port = ntohs(cli_sa.sin_port);
	ONVERBOSE(syslog(LOG_DEBUG, "TCP connection from %s:%d", inet_ntoa(cli_sa.sin_addr), cli_port));

	/*
	 * Connect to destination.
	 */
	struct ip_addr ip;
	ip.addr = (char *) &(cli_sa.sin_addr.s_addr);
	ip.len  = addr_len;
	  
	int rsd = tcp_match(map_list, &ip, cli_port);
	if (rsd == -1) {
	  ONVERBOSE(syslog(LOG_DEBUG, "Address miss"));
	  m_close(csd);
	}
	else {
	  ONVERBOSE(syslog(LOG_DEBUG, "Address match"));

	  if ((csd >= MAX_FD) || (rsd >= MAX_FD)) {
	    syslog(LOG_ERR, "Destination socket descriptors overflow");
	    m_close(csd);
	    m_close(rsd);
	  }
	  else {
	    /*
	     * Add pair of communicating sockets.
	     */
	    FD_SET(csd, &fds);
	    FD_SET(rsd, &fds);
	    maxfd = MAX(maxfd, MAX(rsd + 1, csd + 1)); 
	  
	    /*
	     * Save peers so they can be remembered later.
	     */
	    dest_fd[csd] = rsd;
	    dest_fd[rsd] = csd;
	  }
	}
      }

    } /* sd (mother socket) handled */

    /*
     * Now all other sockets.
     */
    for (int src_fd = 0; nd; ++src_fd)
      if (FD_ISSET(src_fd, &tmp_fds)) {
	--nd;
	int trg_fd = dest_fd[src_fd];
	/*
	 * Copy data.
	 */
	int fail = buf_copy(src_fd, trg_fd, pasv);
	if (fail) {
	  /*
	   * Remove pair of communicating sockets.
	   */
	  FD_CLR(src_fd, &fds);
	  FD_CLR(trg_fd, &fds);
	  m_close(src_fd);
	  m_close(trg_fd);
	  gc_clean(src_fd);
	  gc_clean(trg_fd);
	}
      }

  } /* main loop */

}

void tcp_forward(int port, vector<host_map*> *map_list, int pasv, int uid, int gid)
{
    struct ip_addr local;
    uint32_t addr = INADDR_ANY;
    local.len  = addr_len;
    local.addr = (char *) &addr;
    int sd = tcp_listen(&local, &port, 3);
    if (sd == -1)
      return;
    do_tcp_forward(sd, port, map_list, pasv, uid, gid);
}

void do_udp_forward(struct sockaddr_in *sa, vector<host_map*> *map_list, const char *buf, int buf_len)
{
  int                port      = ntohs(sa->sin_port);
  struct ip_addr     ip;

  ONVERBOSE(syslog(LOG_DEBUG, "TCP packet from: %s:%d\n", 
		   inet_ntoa(sa->sin_addr), 
		   ntohs(sa->sin_port)));

  ip.addr = (char *) &(sa->sin_addr.s_addr);
  ip.len  = addr_len;

  iterator<vector<host_map*>,host_map*> it(*map_list);
  for (it.start(); it.cont(); it.next())
    if (it.get()->udp_match(&ip, port, buf, buf_len))
      break;
}

void udp_forward(int port, vector<host_map*> *map_list, int uid, int gid)
{
  int sd = socket(AF_INET, SOCK_DGRAM, get_protonumber(P_UDP));
  if (sd == -1) {
    syslog(LOG_ERR, "Can't create UDP socket: %s\n", strerror(errno));
    return;
  }
    
  struct sockaddr_in sa;
  sa.sin_family      = AF_INET;
  sa.sin_port        = htons(port);
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  memset((char *) sa.sin_zero, 0, sizeof(sa.sin_zero));

  if (bind(sd, (struct sockaddr *) &sa, sizeof(sa))) {
    syslog(LOG_ERR, "Can't bind UDP socket: %s: %s:%d\n", strerror(errno), inet_ntoa(sa.sin_addr), port);
    close(sd);
    return;
  }

  if (drop_privileges(uid, gid)) {
    close(sd);
    return;
  }

  char buf[BUF_SZ];
  int  rd;
  struct sockaddr_in cli_sa;
  unsigned int cli_sa_len = sizeof(cli_sa);

  for (;;) { /* forever */
    ONVERBOSE(syslog(LOG_DEBUG, "Waiting UDP packet on %s:%d", inet_ntoa(cli_sa.sin_addr), port));

    rd = recvfrom(sd, buf, BUF_SZ, 0, (struct sockaddr *) &cli_sa, &cli_sa_len);
    if (rd == -1) {
      syslog(LOG_ERR, "Can't receive UDP packet: %m");
      break;
    }

    do_udp_forward(&cli_sa, map_list, buf, rd);
  }

  if (close(sd))
    syslog(LOG_WARNING, "close on socket failed: %m");
}

